// Game script and player script that provides a HUD and stamina/sprint functionality to the player.

// ---------------------------------------------------------------------------------------------------------------
// HUD constants

// 32-bit ARGB modulation color (in hex) for base/frame elements, alpha = 102, RGB = 255 255 255
const c_u4HudBaseARGB = 0x66FFFFFF;

// health bar color and alpha
const c_u4HHealthBarARGB = 0xCCD4BE20;
// stamina bar color and alpha
const c_u4HStamBarARGB = 0xCC009BF4;

// pos and dimensions of the bars frame in bars.tga
const c_iBarsFrame_SrcX = 5;
const c_iBarsFrame_SrcY = 4;
const c_iBarsFrame_SrcW = 246;
const c_iBarsFrame_SrcH = 33;
// pos and dims of the filled bar graphic in bars.tga
const c_iBar_SrcX = 37;
const c_iBar_SrcY = 49;
const c_iBar_SrcW = 207;
const c_iBar_SrcH = 8;
// pos, dimensions and offset inside frame of the stamina symbol in bars.tga (used to make it pulsate when sprint key is pressed)
const c_iStamIcon_SrcX = 14;
const c_iStamIcon_SrcY = 21;
const c_iStamIcon_SrcW = 11;
const c_iStamIcon_SrcH = 10;
const c_iStamIcon_OffsetX = 8;
const c_iStamIcon_OffsetY = 16;

// offset for filled health bar inside bars frame
const c_iHealthBar_OffsetX = 32;
const c_iHealthBar_OffsetY = 5;
// offset for filled stamina bar inside bars frame
const c_iStamBar_OffsetX = 28;
const c_iStamBar_OffsetY = 19;

// pos and dimensions of the frame in ammo.tga
const c_iAmmoFrame_SrcX = 5;
const c_iAmmoFrame_SrcY = 4;
const c_iAmmoFrame_SrcW = 54;
const c_iAmmoFrame_SrcH = 33;

// character dimension in digits.tga
const c_iDigitW = 24;
const c_iDigitH = 27;
// character width during output (slightly horizontally compressed to fit in frame)
const c_iDigitOutW = 19;
const c_iDigitSpacing = 15;

// ---------------------------------------------------------------------------------------------------------------
// Player constants

// units per second for stamina to regenerate with
const c_fStaminaRegenRate = 10.0;

// how many stamina points a jump consumes
const c_fStaminaCostForJump = 25.0;

// cost of stamina points per second while sprinting
const c_fStaminaSprintRate = 15.0;

// number of seconds to delay after stamina was consumed before starting auto-regen
const c_fStaminaRegenDelay = 2.0;

// ---------------------------------------------------------------------------------------------------------------
// Game constants

// sprint key uses the USER1 general purpose key bind
uCMD_SPRINT <- uCMD_USER1;


// ---------------------------------------------------------------------------------------------------------------
// custom player class supplying stamina functionality
//
// TScript:
//    string Script = "CPlayerTI"
//
class CPlayerTI extends CPlayer
{
	nvram = null;

	/*********************************************************************************************
	---------------------------------------------------------
	NVRAM vars (vars that need to be preserved in savegames):
	---------------------------------------------------------
	
	// stamina value 0 - 100
	nvram.fStamina

	// a delay timer the prevents stamina regen (when it reaches 0 then regen is allowed)
	nvram.fStaminaRegenDelayTimer

	*********************************************************************************************/

	// true or false depending on if the sprint key is pressed or not
	bSprinting = null;


	function OnScriptInit(bSaveGame)
	{
		if (!bSaveGame)
		{
			// not loading a savegame, init default vals

			// create nvram table
			nvram = {};

			// create and init nvram vars
			nvram.fStamina <- 100.0
			nvram.fStaminaRegenDelayTimer <- 0.0;

			// make sure OnSimStep gets called
			ActivateSimStep(true);
		}

		// init non-saved vars
		bSprinting = false;
	}

	function OnSimStep(fStepDeltaSeconds)
	{
		// handle sprinting
		if (bSprinting)
		{
			ConsumeStamina(c_fStaminaSprintRate * fStepDeltaSeconds);
		}

		// update stamina regen delay timer
		if (nvram.fStaminaRegenDelayTimer > 0.0)
		{
			nvram.fStaminaRegenDelayTimer -= fStepDeltaSeconds;
			if (nvram.fStaminaRegenDelayTimer < 0.0)
				nvram.fStaminaRegenDelayTimer = 0.0;
		}

		// regenerate stamina
		if (nvram.fStamina < 100.0 && nvram.fStaminaRegenDelayTimer == 0.0)
		{
			nvram.fStamina += c_fStaminaRegenRate * fStepDeltaSeconds;
			if (nvram.fStamina > 100.0)
				nvram.fStamina = 100.0;
		}
	}

	// called when the player has been requested to jump, a return value of 1.0 results in standard jump height/force
	// (note that when this function gets called it doesn't mean that the player will definitely end up jumping)
	function fGetJumpFactor()
	{
		if (nvram.fStamina >= c_fStaminaCostForJump)
			// enough stamina for a jump
			return 1.0;
		else
			// no jump for you!
			return 0.0;
	}

	// called when setting player movement speed, fSpeedFactor is the unmodified speed factor
	// the speed factor is 0.5 for walk speed, around 0.78 for walk-strafe combo and 1.0 for run(/max) speed
	function fGetSpeedFactor(fSpeedFactor)
	{
		if (bSprinting && nvram.fStamina > 0.0)
		{
			// when sprinting always moves at maximum sprint speed, regardless of whether player is using the walk or run key
			return 1.75;
		}

		// reduce overall speed when out of stamina
		if (nvram.fStamina == 0.0)
		{
			return fSpeedFactor * 0.85;
		}

		return fSpeedFactor;
	}

	// called when jump is performed
	function OnJump()
	{
		ConsumeStamina(c_fStaminaCostForJump);
	}

	// call this whenever an action consumes stamina
	function ConsumeStamina(amount)
	{
		if (nvram.fStamina > 0.0)
		{
			local fPrevStamin = nvram.fStamina;

			nvram.fStamina -= amount;
			if (nvram.fStamina <= 0.0)
			{
				// just ran out of stamina
				nvram.fStamina = 0.0;
				// TODO: play some sound?
			}
			/*else if (nvram.fStamina < 50.0 && fPrevStamin >= 50.0)
			{
				// stamina just dropped below 50
				// TODO: play some sound ?
			}*/
		}

		// each time stamina is consumed there's a delay before starting regen
		nvram.fStaminaRegenDelayTimer = c_fStaminaRegenDelay;
	}

	// called by game script during input handling to notify this script of the sprint key state
	function UpdateSprintState(bNewState)
	{
		if (bSprinting == bNewState)
			// nothing changed
			return;

		bSprinting = bNewState;

		// TODO: could play some sound when start/stopping sprinting
	}

	function GetStamina()
	{
		return nvram.fStamina;
	}
}


// ---------------------------------------------------------------------------------------------------------------
// helper class that handles HUD drawing for HW rendering (OnPaintHW3D)
class HUD_TI
{
	texHealth = null;
	texAmmo = null;
	texDigits = null;


	constructor()
	{
		// init required resources
		texHealth = TP_Engine.GetTexture("scripts/bars.tga");
		texAmmo = TP_Engine.GetTexture("scripts/ammo.tga");
		texDigits = TP_Engine.GetTexture("scripts/digits.tga");
	}

	// draw a number in the range 0 to 999
	function DrawNumber(iNumber, xpos, ypos, bRightAlign = false)
	{
		// output width and height for one digit
		local dw = c_iDigitOutW;
		local dh = c_iDigitH;

		if (iNumber > 999)
			iNumber = 999;
		else if (iNumber < 0)
			iNumber = 0;

		local s = iNumber.tostring();
		if (bRightAlign)
			xpos -= c_iDigitSpacing * s.len();

		foreach (c in s)
		{
			local i = c - '0';
			TP_Engine.DrawTexture(texDigits,  i * c_iDigitW, 0, c_iDigitW, c_iDigitH,  xpos, ypos, dw, dh,  c_u4HudBaseARGB, 1);
			xpos += c_iDigitSpacing;
		}
	}

	function Draw()
	{
		local plr = TP_Engine.GetPlayer();
		if (!plr)
			return;

		local sw = TP_Engine.GetScreenWidth();
		local sh = TP_Engine.GetScreenHeight();

		const xmargin = 30;
		const ymargin = 30;

		// --------------
		// health display
		// --------------
		if (texHealth)
		{
			// top-left corner of bars frame on screen
			local x = xmargin;
			local y = sh - c_iBarsFrame_SrcH - ymargin;

			// draw frame
			TP_Engine.DrawTexture(
				texHealth,
				c_iBarsFrame_SrcX, c_iBarsFrame_SrcY, c_iBarsFrame_SrcW, c_iBarsFrame_SrcH,
				x, y, c_iBarsFrame_SrcW, c_iBarsFrame_SrcH,
				c_u4HudBaseARGB, 1);

			// draw health bar

			// bar width based on health amount
			local w = (c_iBar_SrcW * (plr.GetHealth() / 100.0)).tointeger();

			TP_Engine.DrawTexture(
				texHealth,
				c_iBar_SrcX, c_iBar_SrcY, w, c_iBar_SrcH,
				x+c_iHealthBar_OffsetX, y+c_iHealthBar_OffsetY, w, c_iBar_SrcH,
				c_u4HHealthBarARGB, 1);

			// draw stamina bar

			// get stamina level from player (requires the player to use the custom CPlayerTI script class with stamina support)
			if (plr instanceof CPlayerTI)
				w = (c_iBar_SrcW * (plr.GetStamina() / 100.0)).tointeger();
			else
				// player doesn't have the right script class, just draw a full stamina bar
				w = c_iBar_SrcW;

			TP_Engine.DrawTexture(
				texHealth,
				c_iBar_SrcX, c_iBar_SrcY, w, c_iBar_SrcH,
				x+c_iStamBar_OffsetX, y+c_iStamBar_OffsetY, w, c_iBar_SrcH,
				c_u4HStamBarARGB, 1);

			// flash stamina icon while sprint key is pressed
			if (plr instanceof CPlayerTI && plr.bSprinting)
			{
				// to flash the icon (with a glow) just take a sine based on sim time to calculate an alpha value
				local fFlashFactor = sin(TP_Engine.GetElapsedSimTime() * 6.28) * 0.5 + 0.5;
				local u4FlashColor = TP_Engine.rgba(255, 255, 255, (255.0 * fFlashFactor).tointeger());

				// glow is drawn with additive/burn blend mode
				TP_Engine.DrawTexture(
					texHealth,
					c_iStamIcon_SrcX, c_iStamIcon_SrcY, c_iStamIcon_SrcW, c_iStamIcon_SrcH,
					x+c_iStamIcon_OffsetX, y+c_iStamIcon_OffsetY, c_iStamIcon_SrcW, c_iStamIcon_SrcH,
					u4FlashColor, 2);
			}

			// TODO: a flashing health icon when health is dangerously low is left as an exercise to the reader
			// (hint: it's basically a copy of the stamina flash just with different position/dimension constants)
		}

		// --------------
		// ammo display
		// --------------
		if (texAmmo && texDigits)
		{
			// only draw this if player is holding a gun (with a valid max ammo count)
			local obj = plr.pinsHeldObject();
			if (obj instanceof CGun && obj.iMaxAmmo > 0)
			{
				// top-left corner of ammo frame on screen
				local x = xmargin + c_iBarsFrame_SrcW + 4;
				local y = sh - c_iBarsFrame_SrcH - ymargin;

				// draw frame
				TP_Engine.DrawTexture(
					texAmmo,
					c_iAmmoFrame_SrcX, c_iAmmoFrame_SrcY, c_iAmmoFrame_SrcW, c_iAmmoFrame_SrcH,
					x, y, c_iAmmoFrame_SrcW, c_iAmmoFrame_SrcH,
					c_u4HudBaseARGB, 1);

				// top-right corner for right-aligned ammo count text
				local tx = c_iAmmoFrame_SrcW - 8;
				local ty = (c_iAmmoFrame_SrcH - c_iDigitH) / 2;

				// draw ammo count (right-aligned)
				DrawNumber(obj.iAmmo, x+tx, y+ty, true);
			}
		}
	}
}


// ---------------------------------------------------------------------------------------------------------------
// A game script that provides custom HUD and other extended game mechanics
class GameTI extends GameScript
{
	hud = null;


	function OnScriptInit(bSaveGame)
	{
		// no HUD provided if game is running with software rendering (too limiting or requires uglier fallback)
		if ( !TP_Engine.IsSoftwareMode() )
			hud = HUD_TI();
	}

	function OnPaintHW3D()
	{
		if (hud)
			hud.Draw();
	}

	function OnHandleInput()
	{
		// handle sprint key (requires player to use the custom CPlayerTI script class)
		local plr = TP_Engine.GetPlayer();
		if (plr && plr instanceof CPlayerTI)
		{
			plr.UpdateSprintState(!plr.bDead() && bKeyPressed(uCMD_SPRINT));
		}
	}
}
